"""Top-level package for IceNet."""

__author__ = "British Antarctic Survey / Alan Turing Institute"
__copyright__ = "British Antarctic Survey"
__email__ = "jambyr@bas.ac.uk"
__license__ = "MIT"
__version__ = "0.2.6"
